from typing import ClassVar, Mapping

import pandas as pd

from ixmp4.data import abstract

from .. import base


class TimeSeries(base.BaseModel):
    NotFound: ClassVar = abstract.TimeSeries.NotFound
    NotUnique: ClassVar = abstract.TimeSeries.NotUnique
    DeletionPrevented: ClassVar = abstract.TimeSeries.DeletionPrevented

    id: int
    run__id: int
    parameters: Mapping


class TimeSeriesRepository(
    base.Creator[TimeSeries],
    base.Retriever[TimeSeries],
    base.Enumerator[TimeSeries],
    base.BulkUpserter[TimeSeries],
    abstract.TimeSeriesRepository,
):
    model_class = TimeSeries
    prefix = "iamc/timeseries/"

    def create(self, run_id: int, parameters: Mapping) -> TimeSeries:
        return super().create(run_id=run_id, parameters=parameters)

    def get(self, run_id: int, parameters: Mapping) -> TimeSeries:
        return super().get(run_ids=[run_id], parameters=parameters)

    def enumerate(self, **kwargs) -> list[TimeSeries] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(self, **kwargs) -> list[TimeSeries]:
        return super()._list(json=kwargs)

    def tabulate(self, join_parameters: bool | None = None, **kwargs) -> pd.DataFrame:
        return super()._tabulate(
            json=kwargs, params={"join_parameters": join_parameters}
        )

    def get_by_id(self, id: int) -> TimeSeries:
        res = self._get_by_id(id)
        return TimeSeries(**res)

    def bulk_upsert(self, df: pd.DataFrame, create_related: bool = False) -> None:
        super().bulk_upsert(df, create_related=create_related)



# type: ignore
"""Create optimization_indexset table

Revision ID: da1fba23f206
Revises: c71efc396d2b
Create Date: 2023-08-30 12:29:50.037072

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# Revision identifiers, used by Alembic.
revision = "da1fba23f206"
down_revision = "c71efc396d2b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "optimization_indexset",
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.Column(
            "elements",
            sa.JSON().with_variant(
                postgresql.JSONB(astext_type=sa.Text()), "postgresql"
            ),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("created_by", sa.String(length=255), nullable=True),
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["run__id"],
            ["run.id"],
            name=op.f("fk_optimization_indexset_run__id_run"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_indexset")),
        sa.UniqueConstraint("name", name=op.f("uq_optimization_indexset_name")),
    )
    with op.batch_alter_table("optimization_indexset", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_optimization_indexset_run__id"),
            ["run__id"],
            unique=False,
        )

    op.create_table(
        "optimization_indexset_docs",
        sa.Column("description", sa.Text(), nullable=False),
        sa.Column("dimension__id", sa.Integer(), nullable=True),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["dimension__id"],
            ["optimization_indexset.id"],
            name=op.f(
                "fk_optimization_indexset_docs_dimension__id_optimization_indexset"
            ),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_optimization_indexset_docs")),
        sa.UniqueConstraint(
            "dimension__id",
            name=op.f("uq_optimization_indexset_docs_dimension__id"),
        ),
    )
    with op.batch_alter_table("iamc_datapoint_oracle", schema=None) as batch_op:
        batch_op.drop_index("ix_iamc_datapoint_oracle_step_category")
        batch_op.drop_index("ix_iamc_datapoint_oracle_step_datetime")
        batch_op.drop_index("ix_iamc_datapoint_oracle_step_year")
        batch_op.drop_index("ix_iamc_datapoint_oracle_time_series__id")
        batch_op.drop_index("ix_iamc_datapoint_oracle_type")

    op.drop_table("iamc_datapoint_oracle")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "iamc_datapoint_oracle",
        sa.Column("value", sa.FLOAT(), nullable=True),
        sa.Column("type", sa.VARCHAR(length=255), nullable=False),
        sa.Column("step_category", sa.VARCHAR(length=1023), nullable=True),
        sa.Column("step_year", sa.INTEGER(), nullable=True),
        sa.Column("step_datetime", sa.DATETIME(), nullable=True),
        sa.Column("time_series__id", sa.INTEGER(), nullable=False),
        sa.Column("id", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["time_series__id"],
            ["iamc_timeseries.id"],
            name="fk_iamc_datapoint_oracle_time_series__id_iamc_timeseries",
        ),
        sa.PrimaryKeyConstraint("id", name="pk_iamc_datapoint_oracle"),
        sa.UniqueConstraint(
            "time_series__id",
            "step_year",
            "step_category",
            "step_datetime",
            name="uq_iamc_datapoint_oracle_time_series__id_step_year_step_category_step_datetime",
        ),
    )
    with op.batch_alter_table("iamc_datapoint_oracle", schema=None) as batch_op:
        batch_op.create_index("ix_iamc_datapoint_oracle_type", ["type"], unique=False)
        batch_op.create_index(
            "ix_iamc_datapoint_oracle_time_series__id",
            ["time_series__id"],
            unique=False,
        )
        batch_op.create_index(
            "ix_iamc_datapoint_oracle_step_year", ["step_year"], unique=False
        )
        batch_op.create_index(
            "ix_iamc_datapoint_oracle_step_datetime",
            ["step_datetime"],
            unique=False,
        )
        batch_op.create_index(
            "ix_iamc_datapoint_oracle_step_category",
            ["step_category"],
            unique=False,
        )

    op.drop_table("optimization_indexset_docs")
    with op.batch_alter_table("optimization_indexset", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_optimization_indexset_run__id"))

    op.drop_table("optimization_indexset")
    # ### end Alembic commands ###



import enum
import logging
import os
import re
from functools import lru_cache

import httpx
import pandas as pd
from pydantic import Field

from ixmp4.core.exceptions import ManagerApiError, PlatformNotFound

from .auth import BaseAuth
from .base import Config, PlatformInfo
from .user import User

logger = logging.getLogger(__name__)


class hashabledict(dict):
    """Hashable dict type used for caching."""

    def __hash__(self):
        return hash(tuple(sorted(self.items())))


class ManagerPlatformInfo(PlatformInfo):
    id: int
    management_group: int
    access_group: int
    url: str
    name: str = Field(alias="slug")
    notice: str | None = Field(default=None)

    class Accessibilty(str, enum.Enum):
        PUBLIC = "PUBLIC"
        GATED = "GATED"
        PRIVATE = "PRIVATE"

    accessibility: Accessibilty


class ManagerConfig(Config):
    template_pattern = re.compile(r"(\{env\:(\w+)\})")

    def __init__(self, url: str, auth: BaseAuth, remote: bool = False) -> None:
        # TODO: Find the sweet-spot for `maxsize`
        # -> a trade-off between memory usage
        # and load on the management service

        self._cached_request = lru_cache(maxsize=128)(self._uncached_request)
        self.url = url
        self.auth = auth
        self.client = httpx.Client(
            base_url=self.url,
            timeout=10.0,
            http2=True,
            auth=auth,
            follow_redirects=True,
        )
        self.remote = remote

    def expand_dsn(self, dsn: str) -> str:
        for template, var in self.template_pattern.findall(dsn):
            if not var.startswith("IXMP4_"):
                # TODO: logging: warn/error if variable name is not valid
                continue
            try:
                val = os.environ[var]
            except KeyError:
                # TODO: logging: warn/error if variable is not in environment
                continue
            dsn = dsn.replace(template, val)
        return dsn

    def _uncached_request(
        self, method: str, path: str, *args, jti: str | None = None, **kwargs
    ):
        del jti
        # `jti` is only used to affect `@lru_cache`
        # if the token id changes a new cache entry will be created
        # TODO: MU improvement:
        # the old cache entry could be completely invalidated/deleted/collected
        # NOTE: since this cache is not shared amongst processes, it's efficacy
        # declines with the scale of the whole infrastructure unless counteracted
        # with increased cache size / memory usage
        res = self.client.request(method, path, *args, **kwargs)
        if res.status_code != 200:
            raise ManagerApiError(f"[{str(res.status_code)}] {res.text}")
        return res.json()

    def _request(
        self,
        method: str,
        path: str,
        *args,
        params: dict | None = None,
        json: dict | list | tuple | None = None,
        **kwargs,
    ):
        if params is not None:
            params = hashabledict(params)

        if json is not None:
            if isinstance(json, dict):
                json = hashabledict(json)
            else:
                json = tuple(json)

        logger.debug(f"Trying cache: {method} {path} {params} {json}")
        return self._cached_request(
            method, path, *args, params=params, json=json, **kwargs
        )

    def fetch_platforms(self, **kwargs) -> list[ManagerPlatformInfo]:
        json = self._request("GET", "/ixmp4", params={"page_size": -1}, **kwargs)
        return [ManagerPlatformInfo(**c) for c in json["results"]]

    def list_platforms(self, **kwargs) -> list[ManagerPlatformInfo]:
        platforms = self.fetch_platforms(**kwargs)

        for i, p in enumerate(platforms):
            if self.remote:
                platforms[i].dsn = p.url
            else:
                platforms[i].dsn = self.expand_dsn(p.dsn)

        return platforms

    def get_platform(self, key: str, **kwargs) -> ManagerPlatformInfo:
        for p in self.list_platforms(**kwargs):
            if p.name == key:
                return p
        else:
            raise PlatformNotFound(
                f"Platform '{key}' does not exist at {self.url} or "
                "you do not have permission to access this platform."
            )

    def fetch_user_permissions(
        self, user: User, platform: ManagerPlatformInfo, **kwargs
    ) -> pd.DataFrame:
        if not user.is_authenticated:
            return pd.DataFrame(
                [], columns=["id", "instance", "group", "access_type", "model"]
            )
        json = self._request(
            "GET",
            "/modelpermissions",
            params={
                "page_size": -1,
                "instance": platform.id,
                "group__users": user.id,
            },
            **kwargs,
        )
        return pd.DataFrame(
            json["results"], columns=["id", "instance", "group", "access_type", "model"]
        )

    def fetch_group_permissions(
        self, group_id: int, platform: ManagerPlatformInfo, **kwargs
    ) -> pd.DataFrame:
        json = self._request(
            "GET",
            "/modelpermissions",
            params={"page_size": -1, "instance": platform.id, "group": group_id},
            **kwargs,
        )
        return pd.DataFrame(
            json["results"], columns=["id", "instance", "group", "access_type", "model"]
        )


class MockManagerConfig(ManagerConfig):
    def __init__(
        self,
        platforms: list[ManagerPlatformInfo],
        permissions: pd.DataFrame,
    ) -> None:
        self.platforms = platforms
        self.permissions = permissions
        self.remote = False

    def fetch_platforms(self) -> list[ManagerPlatformInfo]:
        return self.platforms

    def fetch_user_permissions(
        self, user: User, platform: ManagerPlatformInfo, **kwargs
    ) -> pd.DataFrame:
        pdf = self.permissions
        return pdf.where(pdf["group"].isin(user.groups)).where(
            pdf["instance"] == platform.id
        )

    def fetch_group_permissions(
        self, group_id: int, platform: ManagerPlatformInfo, **kwargs
    ) -> pd.DataFrame:
        pdf = self.permissions
        return pdf.where(pdf["group"] == group_id).where(pdf["instance"] == platform.id)



from typing import Protocol

import pandas as pd

from ixmp4.data import types

from .. import base


class Measurand(base.BaseModel, Protocol):
    """Measurand data model."""

    variable__id: types.Integer
    "Foreign unique integer id of a variable."
    variable: types.Mapped
    "Associated variable."

    unit__id: types.Integer
    "Foreign unique integer id of a unit."
    unit: types.Mapped
    "Associated unit."

    def __str__(self) -> str:
        return f"<Measurand {self.id}>"


class MeasurandRepository(
    base.Creator,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    def create(self, variable_name: str, unit__id: int) -> Measurand:
        ...

    def get(self, variable_name: str, unit__id: int) -> Measurand:
        ...

    def get_or_create(self, variable_name: str, unit__id: int) -> Measurand:
        try:
            return self.get(variable_name, unit__id)
        except Measurand.NotFound:
            return self.create(variable_name, unit__id)

    def list(
        self,
    ) -> list[Measurand]:
        ...

    def tabulate(
        self,
    ) -> pd.DataFrame:
        ...



from pathlib import Path
from typing import Generator

from ixmp4.conf import settings

database_dir = settings.storage_directory / "databases"


def yield_databases() -> Generator[Path, None, None]:
    """Yields all local sqlite database files."""

    databases = database_dir.glob("*.sqlite3")
    return databases


def get_database_path(name: str) -> Path:
    """Returns a :class:`Path` object for a given sqlite database name.
    Does not check whether or not the file actually exists."""

    file_name = name + ".sqlite3"
    return database_dir / file_name


def get_dsn(database_path: Path) -> str:
    """Returns sqlalchemy-friendly sqlite database URI for a given database name."""

    return database_path.absolute().as_uri().replace("file://", "sqlite:///")


def search_databases(name: str) -> str | None:
    """Returns a database URI if the desired database exists, otherwise returns
    `None`."""

    database_path = get_database_path(name)
    if database_path.exists():
        return get_dsn(database_path)
    else:
        return None



from datetime import datetime
from typing import ClassVar

import pandas as pd

from ixmp4.core.base import BaseFacade, BaseModelFacade
from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import IndexSet as IndexSetModel
from ixmp4.data.abstract import Run


class IndexSet(BaseModelFacade):
    _model: IndexSetModel
    NotFound: ClassVar = IndexSetModel.NotFound
    NotUnique: ClassVar = IndexSetModel.NotUnique

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def elements(self) -> list[int | str]:
        return self._model.elements

    def add(self, elements: int | list[int | str] | str) -> None:
        """Adds elements to an existing IndexSet."""
        self.backend.optimization.indexsets.add_elements(
            indexset_id=self._model.id, elements=elements
        )
        self._model.elements = self.backend.optimization.indexsets.get(
            run_id=self._model.run__id, name=self._model.name
        ).elements

    @property
    def run_id(self) -> int:
        return self._model.run__id

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    @property
    def docs(self):
        try:
            return self.backend.optimization.indexsets.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description):
        if description is None:
            self.backend.optimization.indexsets.docs.delete(self.id)
        else:
            self.backend.optimization.indexsets.docs.set(self.id, description)

    @docs.deleter
    def docs(self):
        try:
            self.backend.optimization.indexsets.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<IndexSet {self.id} name={self.name}>"


class IndexSetRepository(BaseFacade):
    _run: Run

    def __init__(self, _run: Run, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._run = _run

    def create(self, name: str) -> IndexSet:
        indexset = self.backend.optimization.indexsets.create(
            run_id=self._run.id,
            name=name,
        )
        return IndexSet(_backend=self.backend, _model=indexset)

    def get(self, name: str) -> IndexSet:
        indexset = self.backend.optimization.indexsets.get(
            run_id=self._run.id, name=name
        )
        return IndexSet(_backend=self.backend, _model=indexset)

    def list(self, name: str | None = None) -> list[IndexSet]:
        indexsets = self.backend.optimization.indexsets.list(name=name)
        return [
            IndexSet(
                _backend=self.backend,
                _model=i,
            )
            for i in indexsets
        ]

    def tabulate(self, name: str | None = None) -> pd.DataFrame:
        return self.backend.optimization.indexsets.tabulate(name=name)



"""
This module contains sqlalchemy database models and repositories.
"""
# flake8: noqa

from .base import BaseModel, BaseRepository
from .docs import BaseDocsRepository, docs_model
from .iamc import (  # AnnualDataPoint,; SubAnnualDataPoint,; CategoricalDataPoint,
    DataPoint,
    DataPointRepository,
    Measurand,
    MeasurandRepository,
    TimeSeries,
    TimeSeriesRepository,
    UniversalDataPoint,
    Variable,
    VariableRepository,
)
from .meta import RunMetaEntry, RunMetaEntryRepository
from .model import Model, ModelRepository
from .optimization import IndexSet, IndexSetRepository, Scalar, ScalarRepository
from .region import Region, RegionRepository
from .run import Run, RunRepository
from .scenario import Scenario, ScenarioRepository
from .unit import Unit, UnitRepository



from typing import ClassVar

from sqlalchemy.ext.declarative import declared_attr

from ixmp4 import db
from ixmp4.data import abstract

from .. import base


class DataPoint(base.BaseModel):
    NotFound: ClassVar = abstract.DataPoint.NotFound
    NotUnique: ClassVar = abstract.DataPoint.NotUnique
    DeletionPrevented: ClassVar = abstract.DataPoint.DeletionPrevented

    Type: ClassVar = abstract.DataPoint.Type

    __abstract__ = True

    updateable_columns = ["value"]

    @declared_attr
    def time_series__id(cls):
        return db.Column(
            db.Integer, db.ForeignKey("iamc_timeseries.id"), nullable=False, index=True
        )

    value = db.Column(db.Float)

    type = db.Column(db.String(255), nullable=False, index=True)

    step_category = db.Column(db.String(1023), index=True)
    step_year = db.Column(db.Integer, index=True)
    step_datetime = db.Column(db.DateTime, index=True)


class UniversalDataPoint(DataPoint):
    @declared_attr.directive
    def __tablename__(cls) -> str:
        return "iamc_datapoint_universal"

    __table_args__ = (
        db.UniqueConstraint("time_series__id", "step_year", "step_category"),
        db.UniqueConstraint("time_series__id", "step_datetime"),
    )


def get_datapoint_model(session) -> type[DataPoint]:
    return UniversalDataPoint



import re

import pandas as pd

from ixmp4 import db
from ixmp4.conf.manager import ManagerConfig, ManagerPlatformInfo
from ixmp4.conf.user import User
from ixmp4.data.db import Model
from ixmp4.db import utils


class AuthorizationContext(object):
    def __init__(
        self, user: User, manager: ManagerConfig, platform: ManagerPlatformInfo
    ):
        self.user = user
        self.manager = manager
        self.platform = platform

    def tabulate_permissions(self):
        df = self.manager.fetch_user_permissions(
            self.user, self.platform, jti=self.user.jti
        )
        if self.platform.accessibility != ManagerPlatformInfo.Accessibilty.PRIVATE:
            group_df = self.manager.fetch_group_permissions(
                self.platform.access_group, self.platform, jti=self.user.jti
            )
            df = pd.concat([df, group_df])
        df = df.dropna()

        def convert_to_regex(m: str) -> str:
            return re.escape(m).replace("\\*", ".*")

        def convert_to_like(m: str) -> str:
            return m.replace("*", "%").replace("_", "[_]")

        df["regex"] = df["model"].apply(convert_to_regex)
        df["like"] = df["model"].apply(convert_to_like)
        return df

    def apply(self, access_type: str, exc: db.sql.Select) -> db.sql.Select:
        if self.is_managed:
            return exc
        if self.user.is_superuser:
            return exc

        if utils.is_joined(exc, Model):
            perms = self.tabulate_permissions()
            if perms.empty:
                return exc.where(False)  # type: ignore
            if access_type == "edit":
                perms = perms.where(perms["access_type"] == "EDIT").dropna()
            # `*` is used as wildcard in permission logic, replaced by sql-wildcard `%`
            conditions = [Model.name.like(p["like"]) for i, p in perms.iterrows()]
            exc = exc.where(db.or_(*conditions))
        return exc

    def check_access(self, access_type: str, model_name: str) -> bool:
        if self.is_managed:
            return True
        if self.user.is_superuser:
            return True

        perms = self.tabulate_permissions()
        if perms.empty:
            return False
        if access_type == "edit":
            perms = perms.where(perms["access_type"] == "EDIT").dropna()

        regex = "^" + "|".join(perms["regex"]) + "$"
        match = re.match(regex, model_name)
        return match is not None

    @property
    def is_accessible(self) -> bool:
        if self.user.is_superuser:
            return True
        if self.is_managed:
            return True
        elif self.platform.accessibility == ManagerPlatformInfo.Accessibilty.PUBLIC:
            return True
        elif self.platform.accessibility == ManagerPlatformInfo.Accessibilty.GATED:
            return self.user.is_authenticated

        return self.platform.access_group in self.user.groups

    @property
    def is_viewable(self) -> bool:
        if self.user.is_superuser:
            return True
        if self.is_managed:
            return True

        df = self.tabulate_permissions()
        return not df[
            (df["access_type"] == "EDIT") | (df["access_type"] == "VIEW")
        ].empty

    @property
    def is_editable(self) -> bool:
        if self.user.is_superuser:
            return True
        if self.is_managed:
            return True

        df = self.tabulate_permissions()
        return not df[df["access_type"] == "EDIT"].empty

    @property
    def is_managed(self) -> bool:
        if self.user.is_superuser:
            return True
        return self.platform.management_group in self.user.groups



from alembic import command
from alembic.config import Config
from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory
from sqlalchemy import create_engine

alembic_config = Config("alembic.ini")


def get_database_revision(dsn: str) -> str | None:
    """Returns the current revision hash of a given database.
    To do this a connection to the database and a call to alembic are required."""

    engine = create_engine(dsn)
    conn = engine.connect()

    context = MigrationContext.configure(conn)
    return context.get_current_revision()


def get_head_revision() -> str | tuple[str, ...] | None:
    """Returns the revision hash of the newest migration
    in the alembic `versions/` directory."""

    script_directory = ScriptDirectory.from_config(alembic_config)
    return script_directory.as_revision_number("head")


def upgrade_database(dsn: str, revision: str = "head") -> None:
    """Uses alembic to upgrade the given database.
    Optionally a desired revision can be supplied instead
    of the default `head` (newest) revision."""

    alembic_config.set_main_option("sqlalchemy.url", dsn)
    command.upgrade(alembic_config, revision)


def stamp_database(dsn: str, revision: str) -> None:
    """Uses alembic to stamp the given database with the given revision.
    Only touches the `alembic_version` table, nothing else."""

    alembic_config.set_main_option("sqlalchemy.url", dsn)
    command.stamp(alembic_config, revision, purge=True)



from datetime import datetime
from typing import ClassVar, Iterable

import pandas as pd

from ixmp4.core.base import BaseFacade, BaseModelFacade
from ixmp4.core.unit import Unit
from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import Run
from ixmp4.data.abstract import Scalar as ScalarModel


class Scalar(BaseModelFacade):
    _model: ScalarModel
    NotFound: ClassVar = ScalarModel.NotFound
    NotUnique: ClassVar = ScalarModel.NotUnique

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def value(self) -> float:
        """Associated value."""
        return self._model.value

    @value.setter
    def value(self, value: float):
        self._model.value = value
        self.backend.optimization.scalars.update(
            id=self._model.id,
            value=self._model.value,
            unit_id=self._model.unit.id,
        )

    @property
    def unit(self):
        """Associated unit."""
        return self._model.unit

    @unit.setter
    def unit(self, unit: str | Unit):
        if isinstance(unit, Unit):
            unit = unit
        else:
            unit_model = self.backend.units.get(unit)
            unit = Unit(_backend=self.backend, _model=unit_model)
        self._model = self.backend.optimization.scalars.update(
            id=self._model.id,
            value=self._model.value,
            unit_id=unit.id,
        )

    @property
    def run_id(self) -> int:
        return self._model.run__id

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    @property
    def docs(self):
        try:
            return self.backend.optimization.scalars.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description):
        if description is None:
            self.backend.optimization.scalars.docs.delete(self.id)
        else:
            self.backend.optimization.scalars.docs.set(self.id, description)

    @docs.deleter
    def docs(self):
        try:
            self.backend.optimization.scalars.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Scalar {self.id} name={self.name}>"


class ScalarRepository(BaseFacade):
    _run: Run

    def __init__(self, _run: Run, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._run = _run

    def create(self, name: str, value: float, unit: str | Unit | None = None) -> Scalar:
        if isinstance(unit, Unit):
            unit_name = unit.name
        elif isinstance(unit, str):
            unit_name = unit
        else:
            # TODO: provide logging information about None-units being converted
            # to dimensionless
            dimensionless_unit = self.backend.units.create(name="dimensionless")
            unit_name = dimensionless_unit.name

        try:
            model = self.backend.optimization.scalars.create(
                name=name, value=value, unit_name=unit_name, run_id=self._run.id
            )
        except Scalar.NotUnique as e:
            raise Scalar.NotUnique(
                message=f"Scalar '{name}' already exists! Did you mean to call "
                "run.optimization.scalars.update()?"
            ) from e
        return Scalar(_backend=self.backend, _model=model)

    def get(self, name: str) -> Scalar:
        model = self.backend.optimization.scalars.get(run_id=self._run.id, name=name)
        return Scalar(_backend=self.backend, _model=model)

    def list(self, name: str | None = None) -> Iterable[Scalar]:
        scalars = self.backend.optimization.scalars.list(run_id=self._run.id, name=name)
        return [
            Scalar(
                _backend=self.backend,
                _model=i,
            )
            for i in scalars
        ]

    def tabulate(self, name: str | None = None) -> pd.DataFrame:
        return self.backend.optimization.scalars.tabulate(name=name)



"""
Check the CLI help command on how to use it:

.. code:: bash

    ixmp4 --help
    ixmp4 platforms --help
    ixmp4 test --help
    ixmp4 server --help

"""

from typing import Optional

import typer

from ixmp4.cli import platforms
from ixmp4.conf import settings
from ixmp4.conf.auth import ManagerAuth
from ixmp4.core.exceptions import InvalidCredentials

from . import utils

app = typer.Typer()
app.add_typer(platforms.app, name="platforms")

try:
    from . import server

    app.add_typer(server.app, name="server")
except ImportError:
    # No server installed
    pass


@app.command()
def login(
    username: str = typer.Argument(..., help="Your username."),
    password: str = typer.Option(
        ...,
        help="Your password. Will be saved in plain-text.",
        prompt=True,
        hide_input=True,
    ),
):
    try:
        auth = ManagerAuth(username, password, str(settings.manager_url))
        user = auth.get_user()
        utils.good(f"Successfully authenticated as user '{user.username}'.")
        if typer.confirm(
            text=(
                "Are you sure you want to save your credentials in plain-text for "
                "future use?"
            )
        ):
            settings.credentials.set("default", username, password)

    except InvalidCredentials:
        raise typer.BadParameter(
            "The credentials you provided are not valid. (Wrong username or password?)"
        )


@app.command()
def logout():
    if typer.confirm(
        "Are you sure you want to log out and delete locally saved credentials?"
    ):
        settings.credentials.clear("default")


try:
    import pytest

    @app.command(
        context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
    )
    def test(
        ctx: typer.Context,
        with_backend: Optional[bool] = False,
        with_benchmarks: Optional[bool] = False,
        dry: Optional[bool] = False,
    ):
        opts = [
            "--cov-report",
            "xml:.coverage.xml",
            "--cov-report",
            "term",
            "--cov=ixmp4",
            "-rsx",
        ] + ctx.args

        if not with_backend:
            opts += ["--ignore=tests/data"]

        if not with_benchmarks:
            opts += ["--benchmark-skip"]
        else:
            opts += ["--benchmark-group-by=func", "--benchmark-columns=min"]

        if dry:
            utils.echo("pytest " + " ".join(opts))
            raise typer.Exit(0)
        exit_code = pytest.main(opts)
        raise typer.Exit(
            code=exit_code.value
            if isinstance(exit_code, pytest.ExitCode)
            else exit_code
        )

except ImportError:
    # No pytest installed
    pass



from datetime import datetime
from typing import Any, ClassVar, Iterable, Mapping

import pandas as pd

from ixmp4.data import abstract

from .. import base
from ..docs import Docs, DocsRepository
from ..unit import Unit


class Scalar(base.BaseModel):
    NotFound: ClassVar = abstract.Scalar.NotFound
    NotUnique: ClassVar = abstract.Scalar.NotUnique
    DeletionPrevented: ClassVar = abstract.Scalar.DeletionPrevented

    id: int
    name: str
    value: float
    unit: Unit
    unit__id: int
    run__id: int

    created_at: datetime | None
    created_by: str | None


class ScalarDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/optimization/scalars/"


class ScalarRepository(
    base.Creator[Scalar],
    base.Retriever[Scalar],
    base.Enumerator[Scalar],
    abstract.ScalarRepository,
):
    model_class = Scalar
    prefix = "optimization/scalars/"
    enumeration_method = "PATCH"

    def __init__(self, client, *args, **kwargs) -> None:
        super().__init__(client, *args, **kwargs)
        self.docs = ScalarDocsRepository(self.client)

    def create(
        self,
        name: str,
        value: float,
        unit_name: str,
        run_id: int,
    ) -> Scalar:
        return super().create(
            name=name, value=value, unit_name=unit_name, run_id=run_id
        )

    def update(
        self, id: int, value: float | None = None, unit_id: int | None = None
    ) -> Scalar:
        # we can assume this type on update endpoints
        res: Mapping[str, Any] = self._request(
            "PATCH",
            self.prefix + str(id) + "/",
            json={
                "value": value,
                "unit_id": unit_id,
            },
        )  # type: ignore
        return self.model_class(**res)

    def get(self, run_id: int, name: str) -> Scalar:
        return super().get(run_id=run_id, name=name)

    def get_by_id(self, id: int) -> Scalar:
        res = self._get_by_id(id)
        return Scalar(**res)

    def list(self, *args, **kwargs) -> Iterable[Scalar]:
        return super().list(*args, **kwargs)

    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)

    def enumerate(self, *args, **kwargs) -> Iterable[Scalar] | pd.DataFrame:
        return super().enumerate(*args, **kwargs)



import logging
import sqlite3
from datetime import datetime, timezone
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Generic,
    Iterable,
    Iterator,
    Tuple,
    TypeVar,
)

import dask.dataframe as dd
import numpy as np
import pandas as pd
from sqlalchemy import event, text
from sqlalchemy.engine import Engine
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.exc import IntegrityError, NoResultFound
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import Bundle, DeclarativeBase, declared_attr
from sqlalchemy.orm.session import Session
from sqlalchemy.sql.schema import Identity, MetaData

from ixmp4 import db
from ixmp4.core.exceptions import Forbidden, IxmpError, ProgrammingError
from ixmp4.data import abstract, types
from ixmp4.db import filters

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

logger = logging.getLogger(__name__)


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    if isinstance(dbapi_connection, sqlite3.Connection):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()


@compiles(Identity, "sqlite")
def visit_identity(element, compiler, **kwargs):
    return text("")


class BaseModel(DeclarativeBase):
    NotFound: ClassVar[type[IxmpError]]
    NotUnique: ClassVar[type[IxmpError]]
    DeletionPrevented: ClassVar[type[IxmpError]]

    table_prefix: str = ""
    updateable_columns: ClassVar[list[str]] = []

    @declared_attr.directive
    def __tablename__(cls: "BaseModel") -> str:
        return str(cls.table_prefix + cls.__name__.lower())

    id: types.Integer = db.Column(
        db.Integer,
        Identity(always=False, on_null=True, start=1, increment=1),
        primary_key=True,
        info={"skip_autogenerate": True},
    )

    def __str__(self):
        return self.__class__.__name__


BaseModel.metadata = MetaData(
    naming_convention={
        "ix": "ix_%(column_0_label)s",
        "uq": "uq_%(table_name)s_%(column_0_N_name)s",
        "ck": "ck_%(table_name)s_%(constraint_name)s",
        "fk": "fk_%(table_name)s_%(column_0_N_name)s_%(referred_table_name)s",
        "pk": "pk_%(table_name)s",
    }
)


ModelType = TypeVar("ModelType", bound=BaseModel)


class BaseRepository(Generic[ModelType]):
    backend: "SqlAlchemyBackend"
    session: Session
    dialect: Dialect
    bundle: Bundle
    model_class: type[ModelType]

    def __init__(self, backend: "SqlAlchemyBackend", *args, **kwargs) -> None:
        self.backend = backend
        self.session = backend.session
        self.engine = backend.engine

        if self.session.bind is not None:
            self.dialect = self.session.bind.dialect
        else:
            raise ProgrammingError("Database session is closed.")

        self.bundle: Bundle = Bundle(
            self.model_class.__name__, *db.utils.get_columns(self.model_class).values()
        )
        super().__init__(*args, **kwargs)


class Retriever(BaseRepository[ModelType], abstract.Retriever):
    def get(self, *args, **kwargs) -> ModelType:
        raise NotImplementedError


class Creator(BaseRepository[ModelType], abstract.Creator):
    def get_creation_info(self) -> dict:
        info = {
            "created_at": datetime.now(tz=timezone.utc),
            "created_by": "@unknown",
        }
        if self.backend.auth_context is not None:
            info["created_by"] = self.backend.auth_context.user.username
        return info

    def add(self, *args, **kwargs) -> ModelType:
        raise NotImplementedError

    def create(self, *args, **kwargs) -> ModelType:
        model = self.add(*args, **kwargs)
        try:
            self.session.commit()
        except IntegrityError:
            self.session.rollback()
            raise self.model_class.NotUnique(*args)

        return model


class Deleter(BaseRepository[ModelType]):
    def delete(self, id: int):
        exc: db.sql.Delete = db.delete(self.model_class).where(
            self.model_class.id == id
        )

        try:
            self.session.execute(
                exc, execution_options={"synchronize_session": "fetch"}
            )
            self.session.commit()
        except NoResultFound:
            raise self.model_class.NotFound
        except IntegrityError:
            raise self.model_class.DeletionPrevented


class Selecter(BaseRepository[ModelType]):
    filter_class: type[filters.BaseFilter]

    def check_access(self, ids: set[int], access_type: str = "view", **kwargs):
        exc = self.select(
            _exc=db.select(db.func.count()).select_from(self.model_class),
            id__in=ids,
            _access_type=access_type,
            **kwargs,
        )
        num_permitted_ids = self.session.execute(exc).scalar()
        num_ids = len(ids)
        if not num_permitted_ids == num_ids:
            logger.debug(
                f"Permission check failed {num_permitted_ids}/{num_ids} objects "
                "permitted."
            )
            raise Forbidden(f"Permission check failed for access type '{access_type}'.")

    def join_auth(self, exc: db.sql.Select) -> db.sql.Select:
        return exc

    def apply_auth(self, exc: db.sql.Select, access_type: str) -> db.sql.Select:
        if self.backend.auth_context is not None:
            if not self.backend.auth_context.is_managed:
                exc = self.join_auth(exc)
            exc = self.backend.auth_context.apply(access_type, exc)
        return exc

    def select(
        self,
        _filter: filters.BaseFilter | None = None,
        _exc: db.sql.Select | None = None,
        _access_type: str = "view",
        _post_filter: Callable[[db.sql.Select], db.sql.Select] | None = None,
        _skip_filter: bool = False,
        **kwargs,
    ) -> db.sql.Select:
        if self.filter_class is None:
            cls_name = self.__class__.__name__
            raise NotImplementedError(
                f"Provide `{cls_name}.filter_class` or reimplement `{cls_name}.select`."
            )

        if _exc is None:
            _exc = db.select(self.model_class)

        _exc = self.apply_auth(_exc, _access_type)

        if _filter is not None and not _skip_filter:
            # for some reason checkers resolve the type of `_filter` to `Unknown`
            filter_instance: filters.BaseFilter = _filter
            _exc = filter_instance.join(_exc, session=self.session)
            _exc = filter_instance.apply(_exc, self.model_class, self.session)
        elif not _skip_filter:
            kwarg_filter = self.filter_class(**kwargs)
            _exc = kwarg_filter.join(_exc, session=self.session)
            _exc = kwarg_filter.apply(_exc, self.model_class, self.session)

        if _post_filter is not None:
            _exc = _post_filter(_exc)
        return _exc


class Lister(Selecter[ModelType]):
    def list(self, *args, **kwargs) -> list[ModelType]:
        _exc = self.select(*args, **kwargs)
        _exc = _exc.order_by(self.model_class.id.asc())
        result = self.session.execute(_exc).scalars().all()
        return list(result)


class Tabulator(Selecter[ModelType]):
    def tabulate(
        self,
        *args,
        _raw: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        _exc = self.select(*args, **kwargs)
        _exc = _exc.order_by(self.model_class.id.asc())

        if self.session.bind is not None:
            with self.engine.connect() as con:
                return pd.read_sql(_exc, con=con).replace([np.nan], [None])
        else:
            raise ProgrammingError("Database session is closed.")


class Enumerator(Lister[ModelType], Tabulator[ModelType]):
    def enumerate(
        self, *args, table: bool = False, **kwargs
    ) -> list[ModelType] | pd.DataFrame:
        if table:
            return self.tabulate(*args, **kwargs)
        else:
            return self.list(*args, **kwargs)

    def paginate(
        self,
        *args,
        limit: int = 1000,
        offset: int = 0,
        **kwargs,
    ) -> list[ModelType] | pd.DataFrame:
        return self.enumerate(
            *args, **kwargs, _post_filter=lambda e: e.offset(offset).limit(limit)
        )

    def count(
        self,
        **kwargs,
    ) -> int:
        _exc = self.select(
            _exc=db.select(db.func.count(self.model_class.id.distinct())).select_from(
                self.model_class
            ),
            **kwargs,
        )
        return self.session.execute(_exc).scalar_one()


class BulkOperator(Tabulator[ModelType]):
    merge_suffix: str = "_y"

    @property
    def max_list_length(self) -> int:
        return 50_000

    def merge_existing(
        self, df: pd.DataFrame, existing_df: pd.DataFrame
    ) -> pd.DataFrame:
        columns = db.utils.get_columns(self.model_class)
        primary_key_columns = db.utils.get_pk_columns(self.model_class)
        foreign_columns = db.utils.get_foreign_columns(self.model_class)
        on = (
            (
                set(existing_df.columns) & set(df.columns) & set(columns.keys())
            )  # all cols which exist in both dfs and the db model
            - set(self.model_class.updateable_columns)  # no updateable columns
            - set(primary_key_columns)  # no pk columns
        )  # = all columns that are constant and provided during creation

        ddf = (
            # https://github.com/dask/dask/issues/9710
            dd.from_pandas(df, chunksize=512_000)  # type: ignore
            .set_index(foreign_columns.keys()[0])
            .merge(
                existing_df,
                how="left",
                on=list(on),
                suffixes=(None, self.merge_suffix),
            )
        )
        return ddf.compute()

    def drop_merge_artifacts(
        self, df: pd.DataFrame, extra_columns: list[str] | None = None
    ) -> pd.DataFrame:
        if extra_columns is None:
            extra_columns = []

        existing_columns = [
            str(col) for col in df.columns if col.endswith(self.merge_suffix)
        ]

        df = df.drop(columns=existing_columns + extra_columns)
        df = df.dropna(axis="columns", how="all")
        df = df.dropna()
        return df

    def split_by_max_unique_values(
        self, df: pd.DataFrame, columns: Iterable[str], mu: int
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        df_len = len(df.index)
        chunk_size = df_len
        remaining_df = pd.DataFrame()

        if chunk_size <= mu:
            return df, remaining_df

        max_ = chunk_size
        chunk_df = df
        while True:
            max_ = max(chunk_df[c].nunique() for c in columns)
            if max_ <= mu:
                break
            chunk_size = int(np.floor((mu / max_) * chunk_size))
            chunk_df = df.iloc[:chunk_size, :]
            remaining_df = df.iloc[chunk_size:, :]

        return chunk_df, remaining_df

    def tabulate_existing(self, df: pd.DataFrame) -> pd.DataFrame:
        exc = db.select(self.model_class)
        foreign_columns = db.utils.get_foreign_columns(self.model_class)

        for col in foreign_columns:
            foreign_pks = df[col.name].unique().tolist()
            exc = exc.where(col.in_(foreign_pks))
        return self.tabulate(_exc=exc, _raw=True, _skip_filter=True)

    def yield_chunks(self, df: pd.DataFrame) -> Iterator[pd.DataFrame]:
        foreign_columns = db.utils.get_foreign_columns(self.model_class)
        foreign_names = [c.name for c in foreign_columns]
        remaining_df = df.sort_values(foreign_names)

        while not remaining_df.empty:
            chunk_df, remaining_df = self.split_by_max_unique_values(
                pd.DataFrame(remaining_df), foreign_names, self.max_list_length
            )
            yield pd.DataFrame(chunk_df)


class BulkUpserter(BulkOperator[ModelType]):
    def bulk_upsert(self, df: pd.DataFrame) -> None:
        # slight performance improvement on small operations
        if len(df.index) < self.max_list_length:
            self.bulk_upsert_chunk(df)
        else:
            for chunk_df in self.yield_chunks(df):
                self.bulk_upsert_chunk(pd.DataFrame(chunk_df))

    def bulk_upsert_chunk(self, df: pd.DataFrame) -> None:
        columns = db.utils.get_columns(self.model_class)
        df = df[list(set(columns.keys()) & set(df.columns))]
        existing_df = self.tabulate_existing(df)
        if existing_df.empty:
            self.bulk_insert(df)
        else:
            df = self.merge_existing(df, existing_df)
            df["exists"] = np.where(pd.notnull(df["id"]), True, False)
            cond = []
            for col in self.model_class.updateable_columns:
                updated_col = col + self.merge_suffix
                if updated_col in df.columns:
                    cond.append(df[col] != df[updated_col])

            df["differs"] = np.where(np.logical_or.reduce(cond), True, False)

            insert_df = self.drop_merge_artifacts(
                df.where(~df["exists"]), extra_columns=["id", "differs", "exists"]
            )
            update_df = self.drop_merge_artifacts(
                df.where(df["exists"] & df["differs"]),
                extra_columns=["differs", "exists"],
            )

            if not insert_df.empty:
                self.bulk_insert(insert_df)
            if not update_df.empty:
                self.bulk_update(update_df)

        self.session.commit()

    def bulk_insert(self, df: pd.DataFrame) -> None:
        # to_dict returns a more general list[Mapping[Hashable, Unknown]]
        m: list[dict[str, Any]] = df.to_dict("records")  # type: ignore

        try:
            self.session.execute(db.insert(self.model_class), m)
        except IntegrityError as e:
            raise self.model_class.NotUnique(*e.args)

    def bulk_update(self, df: pd.DataFrame) -> None:
        # to_dict returns a more general list[Mapping[Hashable, Unknown]]
        m: list[dict[str, Any]] = df.to_dict("records")  # type: ignore
        self.session.bulk_update_mappings(self.model_class, m)  # type: ignore


class BulkDeleter(BulkOperator[ModelType]):
    def bulk_delete(self, df: pd.DataFrame) -> None:
        for chunk_df in self.yield_chunks(df):
            self.bulk_delete_chunk(chunk_df)

    def bulk_delete_chunk(self, df: pd.DataFrame) -> None:
        existing_df = self.tabulate_existing(df)
        if existing_df.empty:
            return

        df = self.merge_existing(df, existing_df)
        df["exists"] = np.where(pd.notnull(df["id"]), True, False)
        delete_df = df.where(df["exists"])
        delete_df = df[["id"]]
        excs = []
        for _, sdf in delete_df.groupby(delete_df.index // self.max_list_length):
            exc = db.delete(self.model_class).where(self.model_class.id.in_(sdf["id"]))
            excs.append(exc)

        for exc in excs:
            self.session.execute(exc, execution_options={"synchronize_session": False})

        self.session.commit()



import re
from itertools import cycle
from pathlib import Path
from typing import Generator, Optional

import typer
from rich.progress import Progress, track

from ixmp4.conf import settings
from ixmp4.conf.manager import ManagerPlatformInfo
from ixmp4.conf.toml import TomlPlatformInfo
from ixmp4.core.exceptions import PlatformNotFound
from ixmp4.core.platform import Platform
from ixmp4.data.generator import MockDataGenerator
from ixmp4.db.utils import alembic, sqlite

from . import utils

app = typer.Typer()


def validate_name(name: str):
    match = re.match(r"^[\w\-_]*$", name)
    if match is None:
        raise typer.BadParameter("Platform name must be slug-like.")
    else:
        return name


def validate_dsn(dsn: str | None):
    if dsn is None:
        return None
    match = re.match(r"^(sqlite|postgresql\+psycopg|https|http)(\:\/\/)", dsn)
    if match is None:
        raise typer.BadParameter(
            "Platform dsn must be a valid URl or database connection string."
        )
    else:
        return dsn


def prompt_sqlite_dsn(name: str):
    path = sqlite.get_database_path(name)
    dsn = sqlite.get_dsn(path)
    if path.exists():
        if typer.confirm(
            f"A file at the standard filesystem location for name '{name}' already "
            "exists. Do you want to add the existing file to the platform registry?"
        ):
            return dsn
        else:
            raise typer.Exit()
    else:
        if typer.confirm(
            f"No file at the standard filesystem location for name '{name}' exists. "
            "Do you want to create a new database?"
        ):
            alembic.upgrade_database(dsn, "head")
            return dsn
        else:
            raise typer.Exit()


@app.command(help="Adds a new platform to ixmp4's toml registry.")
def add(
    name: str = typer.Argument(
        ...,
        help="The string identifier of the platform to add. Must be slug-like.",
        callback=validate_name,
    ),
    dsn: Optional[str] = typer.Option(
        None,
        help="Data source name. Can be a http(s) URL or a database connection string.",
        callback=validate_dsn,
    ),
):
    try:
        settings.toml.get_platform(name)
        raise typer.BadParameter(
            f"Platform with name '{name}' already exists. "
            "Choose another name or remove the existing platform."
        )
    except PlatformNotFound:
        pass

    if dsn is None:
        utils.echo(
            "No DSN supplied, assuming you want to add a local sqlite database..."
        )
        dsn = prompt_sqlite_dsn(name)

    settings.toml.add_platform(name, dsn)
    utils.good("\nPlatform added successfully.")


def prompt_sqlite_removal(dsn: str):
    path = Path(dsn.replace("sqlite://", ""))
    path_str = typer.style(path, fg=typer.colors.CYAN)
    if typer.confirm(
        "Do you want to remove the associated database file at " f"{path_str} as well?"  # type: ignore
    ):
        path.unlink()
        utils.echo("\nDatabase file deleted.")
    else:
        utils.echo("\nDatabase file left intact.")


@app.command(help="Removes a platform from ixmp4's toml registry.")
def remove(
    name: str = typer.Argument(
        ..., help="The string identifier of the platform to remove."
    ),
):
    try:
        platform = settings.toml.get_platform(name)
    except PlatformNotFound:
        raise typer.BadParameter(f"Platform '{name}' does not exist.")

    if typer.confirm(
        f"Are you sure you want to remove the platform '{platform.name}' with dsn "
        f"'{platform.dsn}'?"
    ):
        if platform.dsn.startswith("sqlite://"):
            prompt_sqlite_removal(platform.dsn)
        settings.toml.remove_platform(name)


def tabulate_platforms(
    platforms: list[TomlPlatformInfo] | list[ManagerPlatformInfo],
):
    utils.echo("Platform".ljust(15) + "DSN\n".ljust(15))
    total = 0
    for p in platforms:
        name = p.name
        if len(name) > 12:
            name = name[:9] + "..."
        name = name.ljust(15)

        utils.important(name, nl=False)
        utils.echo(p.dsn.ljust(10))
        total += 1

    utils.info("\n" + str(total), nl=False)
    utils.echo(" total \n")


@app.command("list", help="Lists all registered platforms.")
def list_():
    toml_path_str = typer.style(settings.toml.path, fg=typer.colors.CYAN)
    toml_platforms = settings.toml.list_platforms()
    utils.echo(f"Platforms in '{toml_path_str}':")
    tabulate_platforms(toml_platforms)

    if settings.manager is not None:
        manager_url_str = typer.style(settings.manager.url, fg=typer.colors.CYAN)
        manager_platforms = settings.manager.list_platforms()
        utils.echo(f"Platforms accessible via '{manager_url_str}':")
        tabulate_platforms(manager_platforms)


@app.command(
    help=(
        "Migrates all database platforms from your local toml file to the newest "
        "revision."
    )
)
def upgrade():
    for c in settings.toml.list_platforms():
        if c.dsn.startswith("http"):
            utils.echo(f"Skipping '{c.name}' because it is a REST platform.")
        else:
            utils.echo(f"Upgrading platform '{c.name}' with dsn '{c.dsn}'...")
            alembic.upgrade_database(c.dsn, "head")


@app.command(
    help=(
        "Stamps all database platforms from your local toml file with the given "
        "revision."
    )
)
def stamp(revision: str) -> None:
    for c in settings.toml.list_platforms():
        if c.dsn.startswith("http"):
            utils.echo(f"Skipping '{c.name}' because it is a REST platform.")
        else:
            utils.echo(
                f"Stamping platform '{c.name}' with dsn '{c.dsn}' to '{revision}'..."
            )
            alembic.stamp_database(c.dsn, revision)


@app.command(
    help="Generates mock test data and loads it into a platform. "
    "Experimental and meant as a development tool. Use at your own risk. "
)
def generate(
    platform_name: str,
    num_models: int = typer.Option(
        10, "--models", help="Number of mock models to generate."
    ),
    num_runs: int = typer.Option(40, "--runs", help="Number of mock runs to generate."),
    num_regions: int = typer.Option(
        200, "--regions", help="Number of mock regions to generate."
    ),
    num_variables: int = typer.Option(
        1000, "--variables", help="Number of mock variables to generate."
    ),
    num_units: int = typer.Option(
        40, "--units", help="Number of mock units to generate."
    ),
    num_datapoints: int = typer.Option(
        30_000, "--datapoints", help="Number of mock datapoints to generate."
    ),
):
    try:
        platform = Platform(platform_name)
    except PlatformNotFound:
        raise typer.BadParameter(f"Platform '{platform_name}' does not exist.")

    typer.echo("This command will generate:\n")
    lines = []
    for name, value in [
        ("Model(s)", num_models),
        ("Run(s)", num_runs),
        ("Region(s)", num_regions),
        ("Variable(s)", num_variables),
        ("Units(s)", num_units),
        ("Datapoint(s)", num_datapoints),
    ]:
        value_str = typer.style(str(value), fg=typer.colors.CYAN)
        lines.append(f" - {value_str} {name} ")
    typer.echo("\n".join(lines))
    typer.echo(
        f"...and load them into the platform '{platform_name}' "
        f"(DSN: {platform.backend.info.dsn}).\n"
    )

    if typer.confirm("Are you sure?"):
        generator = MockDataGenerator(
            platform,
            num_models,
            num_runs,
            num_regions,
            num_variables,
            num_units,
            num_datapoints,
        )
        generate_data(generator)
        utils.good("Done!")


def create_cycle(generator: Generator, name: str, total: int):
    return cycle(
        [
            m
            for m in track(
                generator,
                description=f"Generating {name}(s)...",
                total=total,
            )
        ]
    )


def generate_data(generator: MockDataGenerator):
    model_names = create_cycle(
        generator.yield_model_names(), "Model", generator.num_models
    )
    runs = create_cycle(generator.yield_runs(model_names), "Run", generator.num_runs)
    regions = create_cycle(generator.yield_regions(), "Region", generator.num_regions)
    units = create_cycle(generator.yield_units(), "Unit", generator.num_units)
    variable_names = create_cycle(
        generator.yield_variable_names(), "Variable", generator.num_variables
    )
    with Progress() as progress:
        task = progress.add_task(
            description="Generating Datapoint(s)...", total=generator.num_datapoints
        )
        for df in generator.yield_datapoints(runs, variable_names, units, regions):
            progress.advance(task, len(df))



from ixmp4.data.db import filters as base
from ixmp4.db import filters


class OptimizationUnitFilter(base.UnitFilter, metaclass=filters.FilterMeta):
    def join(self, exc, **kwawrgs):
        return exc


class OptimizationScalarFilter(
    base.OptimizationScalarFilter, metaclass=filters.FilterMeta
):
    scalar: base.OptimizationScalarFilter | None
    run: base.RunFilter | None = filters.Field(
        default=base.RunFilter(id=None, version=None, is_default=True)
    )
    unit: base.UnitFilter | None = filters.Field(
        default=OptimizationUnitFilter(id=None, name=None)
    )



import logging
from typing import AsyncGenerator, Callable

import jwt
from fastapi import Depends, Header, Path

from ixmp4.conf import settings
from ixmp4.conf.auth import SelfSignedAuth
from ixmp4.conf.manager import ManagerConfig
from ixmp4.conf.user import User, anonymous_user, local_user
from ixmp4.core.exceptions import Forbidden, InvalidToken, PlatformNotFound
from ixmp4.data.backend.db import SqlAlchemyBackend

logger = logging.getLogger(__name__)
manager = ManagerConfig(
    str(settings.manager_url), SelfSignedAuth(settings.secret_hs256)
)


async def validate_token(authorization: str = Header(None)) -> dict | None:
    """Validates a JSON Web Token with the secret supplied in the
    `IXMP4_SECRET_HS256` environment variable."""

    if authorization is None:
        return None  # anonymous user

    encoded_jwt = authorization.split(" ")[1]
    try:
        return jwt.decode(
            encoded_jwt, settings.secret_hs256, leeway=300, algorithms=["HS256"]
        )
    except jwt.InvalidTokenError as e:
        raise InvalidToken("The supplied token is expired or invalid.") from e


async def do_not_validate_token(authorization: str = Header(None)) -> dict | None:
    """Override dependency used for skipping authentication while testing."""
    return {"user": local_user.model_dump()}


async def get_user(token: dict | None = Depends(validate_token)) -> User:
    """Returns a user object for permission checks."""
    if token is None:
        return anonymous_user

    try:
        user_dict = token["user"]
        user_dict["jti"] = token.get("jti", None)
    except KeyError as e:
        raise InvalidToken("The supplied token is malformed.") from e

    return User(**user_dict)


async def get_version():
    from ixmp4 import __version__

    return __version__


async def get_managed_backend(
    platform: str = Path(), user: User = Depends(get_user)
) -> AsyncGenerator[SqlAlchemyBackend, None]:
    """Returns a platform backend for a platform name as a path parameter.
    Also checks user access permissions if in managed mode."""
    info = manager.get_platform(platform, jti=manager.auth.get_user().jti)

    if info.dsn.startswith("http"):
        raise PlatformNotFound(f"Platform '{platform}' was not found.")
    else:
        backend = SqlAlchemyBackend(info)
        with backend.auth(user, manager, info) as auth:
            if auth.is_accessible:
                yield backend
            else:
                raise Forbidden(
                    f"Access to platform '{platform}' denied "
                    "due to insufficient permissions."
                )
        backend.close()


async def get_toml_backend(
    platform: str = Path(), user: User = Depends(get_user)
) -> AsyncGenerator[SqlAlchemyBackend, None]:
    logger.debug("Looking for platform in `platforms.toml`.")
    info = settings.toml.get_platform(platform)
    if info.dsn.startswith("http"):
        raise PlatformNotFound(f"Platform '{platform}' was not found.")
    else:
        backend = SqlAlchemyBackend(info)
        yield backend
        backend.close()


if settings.managed:
    get_backend = get_managed_backend
else:
    get_backend = get_toml_backend


def get_test_backend_dependency(backend, auth_params) -> Callable:
    async def get_memory_backend(
        platform: str = Path(), user: User = Depends(get_user)
    ) -> AsyncGenerator[SqlAlchemyBackend, None]:
        """Override dependency which always yields a test backend."""
        with backend.auth(*auth_params, overlap_ok=True):
            yield backend

    return get_memory_backend



import pandas as pd

from ixmp4 import db
from ixmp4.data.abstract import iamc as abstract
from ixmp4.data.auth.decorators import guard

from .. import base
from .docs import VariableDocsRepository
from .model import Variable


class VariableRepository(
    base.Creator[Variable],
    base.Retriever[Variable],
    base.Enumerator[Variable],
    abstract.VariableRepository,
):
    model_class = Variable

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.docs = VariableDocsRepository(*args, **kwargs)

        from .filter import VariableFilter

        self.filter_class = VariableFilter

    def add(self, name: str) -> Variable:
        variable = Variable(name=name, **self.get_creation_info())
        self.session.add(variable)
        return variable

    @guard("view")
    def get(self, name: str) -> Variable:
        exc = db.select(Variable).where(Variable.name == name)
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Variable.NotFound

    @guard("edit")
    def create(self, *args, **kwargs) -> Variable:
        return super().create(*args, **kwargs)

    @guard("view")
    def list(self, *args, **kwargs) -> list[Variable]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)



from typing import ClassVar, TypeVar

import pandas as pd
from sqlalchemy.exc import NoResultFound

from ixmp4 import db
from ixmp4.data import abstract, types

from ..auth.decorators import guard
from . import base


class AbstractDocs(base.BaseModel):
    NotFound: ClassVar = abstract.Docs.NotFound
    NotUnique: ClassVar = abstract.Docs.NotUnique
    DeletionPrevented: ClassVar = abstract.Docs.DeletionPrevented

    __abstract__ = True

    description: types.Mapped

    dimension__id: types.Mapped


def docs_model(model: type[base.BaseModel]) -> type[AbstractDocs]:
    dimension__id = db.Column(db.Integer, db.ForeignKey(model.id), unique=True)

    table_dict = {
        "NotFound": AbstractDocs.NotFound,
        "NotUnique": AbstractDocs.NotUnique,
        "__tablename__": model.__tablename__ + "_docs",
        "description": db.Column(db.Text, nullable=False),
        "dimension__id": dimension__id,
    }

    DocsModel = type(model.__name__ + "Docs", (AbstractDocs,), table_dict)
    return DocsModel


DocsType = TypeVar("DocsType", bound=AbstractDocs)


class BaseDocsRepository(
    base.Creator[DocsType],
    base.Retriever[DocsType],
    base.Deleter[DocsType],
    base.Enumerator[DocsType],
    abstract.DocsRepository,
):
    dimension_model_class: ClassVar[type[base.BaseModel]]

    def select(
        self, *, _exc: db.sql.Select | None = None, dimension_id: int | None = None
    ) -> db.sql.Select:
        if _exc is None:
            _exc = db.select(self.model_class)

        if dimension_id is not None:
            _exc = _exc.where(self.model_class.dimension__id == dimension_id)

        return _exc

    def add(self, dimension_id: int, description: str) -> DocsType:
        docs = self.model_class(description=description, dimension__id=dimension_id)
        self.session.add(docs)
        return docs

    @guard("view")
    def get(self, dimension_id: int) -> DocsType:
        exc = self.select(dimension_id=dimension_id)
        try:
            docs = self.session.execute(exc).scalar_one()
            return docs
        except NoResultFound:
            raise self.model_class.NotFound

    @guard("edit")
    def set(self, dimension_id: int, description: str) -> DocsType:
        exc = self.select(dimension_id=dimension_id)
        try:
            docs = self.session.execute(exc).scalar_one()
            docs.description = description
            self.session.commit()
            return docs
        except NoResultFound:
            docs = self.create(dimension_id=dimension_id, description=description)
            return docs

    @guard("edit")
    def delete(self, dimension_id: int) -> None:
        exc: db.sql.Delete = db.delete(self.model_class).where(
            self.model_class.dimension__id == dimension_id
        )

        try:
            self.session.execute(
                exc, execution_options={"synchronize_session": "fetch"}
            )
            self.session.commit()
        # This is weird: if there is no region, we might want to return None
        # for a delete request; but then how do we distinguish a
        # typo also returning None?
        except NoResultFound:
            raise self.model_class.NotFound

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)

    @guard("view")
    def list(self, *args, **kwargs) -> list[DocsType]:
        return super().list(*args, **kwargs)



from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import Variable


class VariableDocsRepository(BaseDocsRepository):
    model_class = docs_model(Variable)  # VariableDocs
    dimension_model_class = Variable



import logging

import httpx
import pandas as pd
from fastapi.testclient import TestClient

from ixmp4.conf import settings
from ixmp4.conf.auth import BaseAuth, SelfSignedAuth, User
from ixmp4.conf.manager import ManagerPlatformInfo, MockManagerConfig, PlatformInfo
from ixmp4.core.exceptions import ImproperlyConfigured, UnknownApiError
from ixmp4.data.api import (
    DataPointRepository,
    IndexSetRepository,
    ModelRepository,
    RunMetaEntryRepository,
    RunRepository,
    ScalarRepository,
    ScenarioRepository,
    TimeSeriesRepository,
    VariableRepository,
)
from ixmp4.data.api.region import RegionRepository
from ixmp4.data.api.unit import UnitRepository
from ixmp4.server import app, v1
from ixmp4.server.rest import APIInfo, deps

from .base import Backend

logger = logging.getLogger(__name__)


class RestBackend(Backend):
    client: httpx.Client

    def __init__(self, info: PlatformInfo, auth: BaseAuth | None = None) -> None:
        super().__init__(info)
        logger.info(f"Connecting to IXMP4 REST API at {info.dsn}")
        if isinstance(info, ManagerPlatformInfo):
            if info.notice is not None:
                logger.info("Platform notice: " + info.notice)
        self.make_client(info.dsn, auth=auth)
        self.create_repositories()

    def make_client(self, rest_url: str, auth: BaseAuth | None):
        auth = self.get_auth(rest_url, auth)

        self.client = httpx.Client(
            base_url=rest_url,
            timeout=30.0,
            http2=True,
            auth=auth,
            follow_redirects=True,
        )

    def get_auth(self, rest_url: str, override_auth: BaseAuth | None) -> BaseAuth:
        root = httpx.get(rest_url, follow_redirects=True)
        if root.status_code != 200:
            logger.error("Root API response not OK: " + root.text)
            raise UnknownApiError(f"Server response not OK. ({root.status_code})")

        api_info = APIInfo(**root.json())
        logger.info(f"Connected to Platform '{api_info.name}'")
        logger.info("Server IXMP4 Version: " + api_info.version)

        logger.debug("Server UTC Time: " + api_info.utcnow.strftime("%c"))
        logger.debug("Server Is Managed: " + str(api_info.is_managed))
        if api_info.manager_url is not None:
            if (
                api_info.manager_url.rstrip("/")
                != str(settings.manager_url).rstrip("/")
                and api_info.is_managed
            ):
                logger.error(f"Server Manager URL: {api_info.manager_url}")
                logger.error(f"Local Manager URL: {settings.manager_url}")
                raise ImproperlyConfigured(
                    "Trying to connect to a managed REST Platform "
                    "with a mismatching Manager URL."
                )

        if override_auth is None:
            if api_info.is_managed:
                return settings.default_auth
            else:
                logger.info(
                    "Connecting to unmanaged server instance, "
                    "falling back to self-signed auth."
                )
                return SelfSignedAuth(settings.secret_hs256)
        else:
            return override_auth

    def create_repositories(self):
        self.iamc.datapoints = DataPointRepository(self.client)
        self.iamc.timeseries = TimeSeriesRepository(self.client)
        self.iamc.variables = VariableRepository(self.client)
        self.meta = RunMetaEntryRepository(self.client)
        self.models = ModelRepository(self.client)
        self.optimization.indexsets = IndexSetRepository(self.client)
        self.optimization.scalars = ScalarRepository(self.client)
        self.regions = RegionRepository(self.client)
        self.runs = RunRepository(self.client)
        self.scenarios = ScenarioRepository(self.client)
        self.units = UnitRepository(self.client)


test_platform = ManagerPlatformInfo(
    id=1,
    management_group=1,
    access_group=1,
    accessibility=ManagerPlatformInfo.Accessibilty.PRIVATE,
    slug="test",
    notice="Welcome to the test platform.",
    dsn="http://testserver/v1/:test:/",
    url="http://testserver/v1/:test:/",
)

test_user = User(id=-1, username="@unknown", is_superuser=True, is_verified=True)

test_permissions = pd.DataFrame(
    [], columns=["id", "instance", "group", "access_type", "model"]
)

mock_manager = MockManagerConfig([test_platform], test_permissions)


class RestTestBackend(RestBackend):
    def __init__(self, db_backend, *args, **kwargs) -> None:
        self.db_backend = db_backend
        self.auth_params = (test_user, mock_manager, test_platform)
        super().__init__(
            test_platform, SelfSignedAuth(settings.secret_hs256), *args, **kwargs
        )

    def make_client(self, rest_url: str, auth: BaseAuth):
        self.client = TestClient(app=app, base_url=rest_url)

        app.dependency_overrides[deps.validate_token] = deps.do_not_validate_token
        v1.dependency_overrides[deps.validate_token] = deps.do_not_validate_token

        app.dependency_overrides[deps.get_backend] = deps.get_test_backend_dependency(
            self.db_backend, self.auth_params
        )
        v1.dependency_overrides[deps.get_backend] = deps.get_test_backend_dependency(
            self.db_backend, self.auth_params
        )

    def close(self):
        self.client.close()



from fastapi import APIRouter, Body, Depends, Query

from ixmp4.core.exceptions import BadRequest
from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.iamc.datapoint.filter import DataPointFilter

from .. import deps
from ..base import EnumerationOutput, Pagination
from ..decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/datapoints",
    tags=["iamc", "datapoints"],
)


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.DataPoint])
def query(
    filter: DataPointFilter = Body(DataPointFilter()),
    join_parameters: bool | None = Query(False),
    join_runs: bool = Query(False),
    table: bool | None = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    """This endpoint is used to retrieve and optionally filter data.add()

    Filter parameters are provided as keyword arguments.

    The available filters can be found here:
    :class:`ixmp4.data.db.iamc.datapoint.filter.DataPointFilter`.

    Examples
    --------

    Filter data points for a given model, scenario combination, and a number of years:

    ..  code-block:: json

        {
            "model" : {"name": "model 1"},
            "scenario" : {"name": "scenario 1"},
            "year__in" : [2020, 2025]
        }

    Return all data for a given variable:

    .. code-block:: json

        {
            "variable": {"name": "Final Energy"}
        }


    """
    if (join_parameters or join_runs) and not table:
        raise BadRequest(
            "`join_parameters` or `join_run` can only be used with `table=true`."
        )

    return EnumerationOutput(
        results=backend.iamc.datapoints.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
            join_parameters=join_parameters,
            join_runs=join_runs,
        ),
        total=backend.iamc.datapoints.count(
            _filter=filter,
            join_parameters=join_parameters,
            join_runs=join_runs,
        ),
        pagination=pagination,
    )


@autodoc
@router.post("/bulk/")
def bulk_upsert(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.datapoints.bulk_upsert(df.to_pandas())


@autodoc
@router.patch("/bulk/")
def bulk_delete(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.datapoints.bulk_delete(df.to_pandas())



# flake8: noqa

from .datapoint import (  # AnnualDataPoint,; SubAnnualDataPoint,; CategoricalDataPoint,
    DataPoint,
    DataPointRepository,
)
from .timeseries import TimeSeries, TimeSeriesRepository
from .variable import Variable, VariableRepository



import functools

import pandera as pa
from pandera.errors import SchemaError as PanderaSchemaError

from .exceptions import SchemaError


def check_types(func):
    checked_func = pa.check_types(func)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return checked_func(*args, **kwargs)
        except PanderaSchemaError as e:
            raise SchemaError(*e.args)

    return wrapper



import logging
from datetime import datetime, timedelta
from uuid import uuid4

import httpx
import jwt

from ixmp4.core.exceptions import InvalidCredentials, IxmpError

from .user import User, anonymous_user

logger = logging.getLogger(__name__)


class BaseAuth(object):
    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def auth_flow(self, request):
        yield self(request)

    def get_user(self) -> User:
        raise NotImplementedError


class SelfSignedAuth(BaseAuth, httpx.Auth):
    """Generates its own JWT with the supplied secret."""

    def __init__(self, secret: str, username: str = "ixmp4"):
        self.secret = secret
        self.user = User(
            id=-1,
            username=username,
            email="ixmp4@iiasa.ac.at",
            is_staff=True,
            is_superuser=True,
            is_verified=True,
            groups=[],
        )
        self.token = self.get_local_jwt()

    def __call__(self, r):
        try:
            jwt.decode(self.token, self.secret, algorithms=["HS256"])
        except (jwt.InvalidTokenError, jwt.ExpiredSignatureError):
            self.token = self.get_local_jwt()

        r.headers["Authorization"] = "Bearer " + self.token
        return r

    def get_local_jwt(self):
        self.jti = uuid4().hex
        return jwt.encode(
            {
                "token_type": "access",
                "exp": self.get_expiration_timestamp(),
                "jti": self.jti,
                "sub": "ixmp4",
                "iss": "ixmp4",
                "user": self.user.model_dump(),
            },
            self.secret,
            algorithm="HS256",
        )

    def get_expiration_timestamp(self):
        return int((datetime.now() + timedelta(minutes=15)).timestamp())

    def get_user(self) -> User:
        self.user.jti = self.jti
        return self.user


class AnonymousAuth(BaseAuth, httpx.Auth):
    def __init__(self):
        self.user = anonymous_user
        logger.info("Connecting to service anonymously and without credentials.")

    def __call__(self, r):
        return r

    def get_user(self) -> User:
        return self.user


class ManagerAuth(BaseAuth, httpx.Auth):
    """Uses the SceSe AAC/Management Service to obtain and refresh a token."""

    def __init__(
        self,
        username: str,
        password: str,
        url: str,
    ):
        self.client = httpx.Client(base_url=url, timeout=10.0, http2=True)
        self.username = username
        self.password = password
        self.obtain_jwt()

    def __call__(self, r):
        try:
            jwt.decode(
                self.access_token,
                options={"verify_signature": False, "verify_exp": True},
            )
        except jwt.ExpiredSignatureError:
            self.refresh_or_reobtain_jwt()

        r.headers["Authorization"] = "Bearer " + self.access_token
        return r

    def obtain_jwt(self):
        res = self.client.post(
            "/token/obtain/",
            json={
                "username": self.username,
                "password": self.password,
            },
        )
        if res.status_code >= 400:
            if res.status_code == 401:
                raise InvalidCredentials(
                    "Your credentials were rejected by the "
                    "Scenario Services Management System. "
                    "Check if they are correct and your account is active "
                    "or log out with `ixmp4 logout` to use ixmp4 anonymously."
                )
            else:
                raise IxmpError("Unknown API error: " + res.text)

        json = res.json()
        self.access_token = json["access"]
        self.set_user(self.access_token)
        self.refresh_token = json["refresh"]

    def refresh_or_reobtain_jwt(self):
        try:
            jwt.decode(
                self.refresh_token,
                options={"verify_signature": False, "verify_exp": True},
            )
            self.refresh_jwt()
        except jwt.ExpiredSignatureError:
            self.obtain_jwt()

    def refresh_jwt(self):
        res = self.client.post(
            "/token/refresh/",
            json={
                "refresh": self.refresh_token,
            },
        )

        if res.status_code >= 400:
            raise IxmpError("Unknown API error: " + res.text)

        self.access_token = res.json()["access"]
        self.set_user(self.access_token)

    def decode_token(self, token: str):
        return jwt.decode(
            token,
            options={"verify_signature": False, "verify_exp": False},
        )

    def set_user(self, token: str):
        token_dict = self.decode_token(token)
        user_dict = token_dict["user"]
        self.user = User(**user_dict, jti=token_dict.get("jti"))

    def get_user(self) -> User:
        return self.user



import pandas as pd
from sqlalchemy.exc import NoResultFound

from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard

from .. import base
from .docs import RegionDocsRepository
from .model import Region


class RegionRepository(
    base.Creator[Region],
    base.Deleter[Region],
    base.Retriever[Region],
    base.Enumerator[Region],
    abstract.RegionRepository,
):
    model_class = Region

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        from .filter import RegionFilter

        self.filter_class = RegionFilter
        self.docs = RegionDocsRepository(*args, **kwargs)

    def add(self, name: str, hierarchy: str) -> Region:
        region = Region(name=name, hierarchy=hierarchy, **self.get_creation_info())
        self.session.add(region)
        return region

    @guard("manage")
    def create(self, *args, **kwargs) -> Region:
        return super().create(*args, **kwargs)

    @guard("manage")
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @guard("view")
    def get(self, name: str) -> Region:
        exc = self.select().where(Region.name == name)
        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise Region.NotFound

    @guard("view")
    def list(self, *args, **kwargs) -> list[Region]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)



# flake8: noqa

from .base import BaseModel, DataFrame
from .docs import Docs, DocsRepository
from .iamc import (  # Measurand,; MeasurandRepository,; AnnualDataPoint,; SubAnnualDataPoint,; CategoricalDataPoint,
    DataPoint,
    DataPointRepository,
    TimeSeries,
    TimeSeriesRepository,
    Variable,
    VariableRepository,
)
from .meta import RunMetaEntry, RunMetaEntryRepository
from .model import Model, ModelRepository
from .optimization import IndexSet, IndexSetRepository, Scalar, ScalarRepository
from .region import Region, RegionParent, RegionRepository
from .run import Run, RunRepository
from .scenario import Scenario, ScenarioRepository
from .unit import Unit, UnitRepository



from ixmp4 import db
from ixmp4.data.db import filters as base
from ixmp4.data.db.iamc.datapoint import get_datapoint_model
from ixmp4.data.db.iamc.measurand import Measurand
from ixmp4.data.db.iamc.timeseries import TimeSeries
from ixmp4.db import filters, utils

from . import Unit


class BaseIamcFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    def join_datapoints(self, exc, session=None):
        if not utils.is_joined(exc, Measurand):
            exc = exc.join(Measurand, Measurand.unit__id == Unit.id)

        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(
                TimeSeries, onclause=TimeSeries.measurand__id == Measurand.id
            )

        model = get_datapoint_model(session)
        if not utils.is_joined(exc, model):
            exc = exc.join(model, onclause=model.time_series__id == TimeSeries.id)
        return exc


class SimpleIamcUnitFilter(
    base.UnitFilter, BaseIamcFilter, metaclass=filters.FilterMeta
):
    def join(self, exc, session=None):
        return super().join_datapoints(exc, session)


class IamcUnitFilter(base.UnitFilter, BaseIamcFilter, metaclass=filters.FilterMeta):
    variable: base.VariableFilter
    region: base.RegionFilter
    run: base.RunFilter = filters.Field(
        default=base.RunFilter(id=None, version=None, is_default=True)
    )

    def join(self, exc, session=None):
        return super().join_datapoints(exc, session)


class UnitFilter(base.UnitFilter, BaseIamcFilter, metaclass=filters.FilterMeta):
    iamc: IamcUnitFilter | filters.Boolean

    def filter_iamc(self, exc, c, v, session=None):
        if v is None:
            return exc

        if v is True:
            return self.join_datapoints(exc, session)
        else:
            ids = self.join_datapoints(db.select(Unit.id), session)
            exc = exc.where(~Unit.id.in_(ids))
            return exc

    def join(self, exc, **kwargs):
        return exc



from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard

from .. import base
from ..docs import BaseDocsRepository, docs_model
from .model import Region

RegionDocs = docs_model(Region)


class RegionDocsRepository(BaseDocsRepository, base.BaseRepository):
    model_class = RegionDocs
    dimension_model_class = Region

    @guard("view")
    def list(self, *, dimension_id: int | None = None) -> list[abstract.Docs]:
        return super().list(dimension_id=dimension_id)



from ixmp4.data.db import filters as base
from ixmp4.data.db.iamc.datapoint import get_datapoint_model
from ixmp4.data.db.iamc.timeseries import TimeSeries
from ixmp4.db import filters, utils

from ..measurand import Measurand
from . import Variable


class VariableFilter(base.VariableFilter, metaclass=filters.FilterMeta):
    region: base.RegionFilter | None
    unit: base.UnitFilter | None
    run: base.RunFilter = filters.Field(
        default=base.RunFilter(id=None, version=None, is_default=True)
    )

    def join(self, exc, session=None):
        if not utils.is_joined(exc, Measurand):
            exc = exc.join(Measurand, Measurand.variable__id == Variable.id)

        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(
                TimeSeries, onclause=TimeSeries.measurand__id == Measurand.id
            )

        model = get_datapoint_model(session)
        if not utils.is_joined(exc, model):
            exc = exc.join(model, onclause=model.time_series__id == TimeSeries.id)
        return exc



from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import Scalar


class ScalarDocsRepository(BaseDocsRepository):
    model_class = docs_model(Scalar)  # ScalarDocs
    dimension_model_class = Scalar



from typing import Any

from fastapi import APIRouter, Body, Depends, Query, Response

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.iamc.timeseries.filter import TimeSeriesFilter

from .. import deps
from ..base import BaseModel, EnumerationOutput, Pagination
from ..decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/timeseries",
    tags=["iamc", "timeseries"],
)


class TimeSeriesInput(BaseModel):
    run__id: int
    parameters: dict[str, Any]


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.TimeSeries])
def query(
    join_parameters: bool | None = Query(False),
    filter: TimeSeriesFilter = Body(TimeSeriesFilter()),
    table: bool | None = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.iamc.timeseries.paginate(
            _filter=filter,
            join_parameters=join_parameters,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.iamc.timeseries.count(_filter=filter),
        pagination=pagination,
    )


@router.post("/", response_model=api.Run)
def create(
    timeseries: TimeSeriesInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.timeseries.create(**timeseries.model_dump())


@router.get("/{id}/", response_model=api.TimeSeries)
def get_by_id(
    id: int,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.timeseries.get_by_id(id)


@router.post("/bulk/")
def bulk_upsert(
    df: api.DataFrame,
    create_related: bool | None = Query(False),
    backend: Backend = Depends(deps.get_backend),
):
    backend.iamc.timeseries.bulk_upsert(df.to_pandas(), create_related=create_related)
    return Response(status_code=201)



from datetime import datetime
from typing import ClassVar

import pandas as pd
from pydantic import Field

from ixmp4.data import abstract

from . import base
from .docs import Docs, DocsRepository


class Scenario(base.BaseModel):
    NotFound: ClassVar = abstract.Scenario.NotFound
    NotUnique: ClassVar = abstract.Scenario.NotUnique
    DeletionPrevented: ClassVar = abstract.Scenario.DeletionPrevented

    id: int
    name: str
    created_at: datetime | None = Field(None)
    created_by: str | None = Field(None)


class ScenarioDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/scenarios/"


class ScenarioRepository(
    base.Creator[Scenario],
    base.Retriever[Scenario],
    base.Enumerator[Scenario],
    abstract.ScenarioRepository,
):
    model_class = Scenario
    prefix = "scenarios/"

    def __init__(self, client, *args, **kwargs) -> None:
        super().__init__(client, *args, **kwargs)
        self.docs = ScenarioDocsRepository(self.client)

    def create(
        self,
        name: str,
    ) -> Scenario:
        return super().create(name=name)

    def get(self, name: str) -> Scenario:
        return super().get(name=name)

    def enumerate(self, **kwargs) -> list[Scenario] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(self, **kwargs) -> list[Scenario]:
        return super()._list(json=kwargs)

    def tabulate(self, **kwargs) -> pd.DataFrame:
        return super()._tabulate(json=kwargs)



from typing import ClassVar

from ixmp4 import db
from ixmp4.data import abstract, types

from .. import base


class Unit(base.BaseModel):
    NotFound: ClassVar = abstract.Unit.NotFound
    NotUnique: ClassVar = abstract.Unit.NotUnique
    DeletionPrevented: ClassVar = abstract.Unit.DeletionPrevented

    name: types.String = db.Column(db.String(255), nullable=False, unique=True)

    created_at = db.Column(db.DateTime)
    created_by = db.Column(db.String(255))



from typing import Iterable

import pandas as pd

from ixmp4 import db
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.auth.decorators import guard

from .. import base
from .docs import ScalarDocsRepository
from .model import Scalar


class ScalarRepository(
    base.Creator[Scalar],
    base.Retriever[Scalar],
    base.Enumerator[Scalar],
    abstract.ScalarRepository,
):
    model_class = Scalar

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.docs = ScalarDocsRepository(*args, **kwargs)

        from .filter import OptimizationScalarFilter

        self.filter_class = OptimizationScalarFilter

    def add(self, name: str, value: float | int, unit_name: str, run_id: int) -> Scalar:
        unit_id = self.backend.units.get(unit_name).id
        scalar = Scalar(
            name=name,
            value=value,
            unit__id=unit_id,
            run__id=run_id,
            **self.get_creation_info(),
        )
        self.session.add(scalar)
        return scalar

    @guard("view")
    def get(self, run_id: int, name: str) -> Scalar:
        exc = db.select(Scalar).where(
            (Scalar.name == name) & (Scalar.run__id == run_id)
        )
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise Scalar.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> Scalar:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Scalar.NotFound(id=id)

        return obj

    @guard("edit")
    def create(
        self, name: str, value: float, unit_name: str, run_id: int, **kwargs
    ) -> Scalar:
        return super().create(
            name=name, value=value, unit_name=unit_name, run_id=run_id, **kwargs
        )

    @guard("edit")
    def update(
        self, id: int, value: float | None = None, unit_id: int | None = None
    ) -> Scalar:
        exc = db.update(Scalar).where(
            Scalar.id == id,
        )

        if value is not None:
            exc = exc.values(value=value)
        if unit_id is not None:
            exc = exc.values(unit__id=unit_id)

        self.session.execute(exc)
        self.session.commit()
        return self.get_by_id(id)

    @guard("view")
    def list(self, *args, **kwargs) -> Iterable[Scalar]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)



from fastapi import APIRouter, Body, Depends, Path, Query
from pydantic import Field

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.run.filter import RunFilter

from . import deps
from .base import BaseModel, EnumerationOutput, Pagination

router: APIRouter = APIRouter(
    prefix="/runs",
    tags=["runs"],
)


class RunInput(BaseModel):
    name_of_model: str = Field(..., alias="model_name")
    scenario_name: str


@router.patch("/", response_model=EnumerationOutput[api.Run])
def query(
    filter: RunFilter = Body(RunFilter()),
    table: bool | None = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.runs.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.runs.count(_filter=filter),
        pagination=pagination,
    )


@router.post("/", response_model=api.Run)
def create(
    run: RunInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.runs.create(**run.model_dump(by_alias=True))


@router.post("/{id}/set-as-default-version/")
def set_as_default_version(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    backend.runs.set_as_default_version(id)


@router.post("/{id}/unset-as-default-version/")
def unset_as_default_version(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    backend.runs.unset_as_default_version(id)



from typing import List

import pandas as pd

from ixmp4 import db
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.auth.decorators import guard

from .. import base
from .docs import IndexSetDocsRepository
from .model import IndexSet


class IndexSetRepository(
    base.Creator[IndexSet],
    base.Retriever[IndexSet],
    base.Enumerator[IndexSet],
    abstract.IndexSetRepository,
):
    model_class = IndexSet

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.docs = IndexSetDocsRepository(*args, **kwargs)

        from .filter import OptimizationIndexSetFilter

        self.filter_class = OptimizationIndexSetFilter

    def add(self, run_id: int, name: str) -> IndexSet:
        indexset = IndexSet(run__id=run_id, name=name, **self.get_creation_info())
        self.session.add(indexset)
        return indexset

    @guard("view")
    def get(self, run_id: int, name: str) -> IndexSet:
        exc = db.select(IndexSet).where(
            (IndexSet.name == name) & (IndexSet.run__id == run_id)
        )
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise IndexSet.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> IndexSet:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise IndexSet.NotFound(id=id)

        return obj

    @guard("edit")
    def create(self, run_id: int, name: str, **kwargs) -> IndexSet:
        return super().create(run_id=run_id, name=name, **kwargs)

    @guard("view")
    def list(self, *args, **kwargs) -> list[IndexSet]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)

    @guard("edit")
    def add_elements(
        self,
        indexset_id: int,
        elements: int | List[int | str] | str,
    ) -> None:
        indexset = self.get_by_id(id=indexset_id)
        if not isinstance(elements, list):
            elements = [elements]
        if indexset.elements is None:
            indexset.elements = elements
        else:
            indexset.elements = indexset.elements + elements

        self.session.add(indexset)
        self.session.commit()



from fastapi import APIRouter, Body, Depends, Path, Query

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.unit.filter import UnitFilter

from . import deps
from .base import BaseModel, EnumerationOutput, Pagination
from .decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/units",
    tags=["units"],
)


class UnitInput(BaseModel):
    name: str


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.Unit])
def query(
    filter: UnitFilter = Body(None),
    table: bool = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.units.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.units.count(_filter=filter),
        pagination=pagination,
    )


@router.post("/", response_model=api.Unit)
def create(
    unit: UnitInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.units.create(**unit.model_dump())


@router.delete("/{id}/")
def delete(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    backend.units.delete(id)



import pydantic


class User(pydantic.BaseModel):
    id: int | None = None
    username: str
    email: str | None = None
    is_staff: bool = False
    is_superuser: bool = False
    is_verified: bool = False
    is_authenticated: bool = True
    groups: list[int] = []
    jti: str | None = None


local_user = User(
    username="@unknown", is_staff=True, is_superuser=True, is_verified=True
)

anonymous_user = User(username="@anonymous", is_authenticated=False)



"""
This module holds a shared datastructure and interface for normalization
between the database and api data models and repositories.
"""
# flake8: noqa

from .base import (
    BaseMeta,
    BaseModel,
    BulkDeleter,
    BulkUpserter,
    Creator,
    Deleter,
    Enumerator,
    Lister,
    Retriever,
    Tabulator,
)
from .docs import Docs, DocsRepository
from .iamc import (  # AnnualDataPoint,; SubAnnualDataPoint,; CategoricalDataPoint,
    DataPoint,
    DataPointRepository,
    Measurand,
    MeasurandRepository,
    TimeSeries,
    TimeSeriesRepository,
    Variable,
    VariableRepository,
)
from .meta import MetaValue, RunMetaEntry, RunMetaEntryRepository, StrictMetaValue
from .model import Model, ModelRepository
from .optimization import IndexSet, IndexSetRepository, Scalar, ScalarRepository
from .region import Region, RegionRepository
from .run import Run, RunRepository
from .scenario import Scenario, ScenarioRepository
from .unit import Unit, UnitRepository



from typing import ClassVar

import pandas as pd
from pydantic import Field

from ixmp4.data import abstract

from . import base
from .model import Model
from .scenario import Scenario


class Run(base.BaseModel):
    NotFound: ClassVar = abstract.Run.NotFound
    NotUnique: ClassVar = abstract.Run.NotUnique
    DeletionPrevented: ClassVar = abstract.Run.DeletionPrevented

    NoDefaultVersion: ClassVar = abstract.Run.NoDefaultVersion

    id: int

    model: Model
    id_of_model: int = Field(..., alias="model__id")

    scenario: Scenario
    scenario__id: int

    version: int
    is_default: bool


class RunRepository(
    base.Creator[Run],
    base.Retriever[Run],
    base.Enumerator[Run],
    abstract.RunRepository,
):
    model_class = Run
    prefix = "runs/"

    def create(self, model_name: str, scenario_name: str) -> Run:
        return super().create(model_name=model_name, scenario_name=scenario_name)

    def get(self, model_name: str, scenario_name: str, version: int) -> Run:
        return super().get(
            model={"name": model_name},
            scenario={"name": scenario_name},
            version=version,
            default_only=False,
            is_default=None,
        )

    def enumerate(self, **kwargs) -> list[Run] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(self, **kwargs) -> list[Run]:
        return super()._list(json=kwargs)

    def tabulate(self, **kwargs) -> pd.DataFrame:
        return super()._tabulate(json=kwargs)

    def get_default_version(self, model_name: str, scenario_name: str) -> Run:
        try:
            return super().get(
                model={"name": model_name},
                scenario={"name": scenario_name},
                is_default=True,
            )
        except Run.NotFound:
            raise Run.NoDefaultVersion

    def set_as_default_version(self, id: int) -> None:
        self._request(
            "POST",
            self.prefix + "/".join([str(id), "set-as-default-version/"]),
        )

    def unset_as_default_version(self, id: int) -> None:
        self._request(
            "POST",
            self.prefix + "/".join([str(id), "unset-as-default-version/"]),
        )



from dotenv import load_dotenv

from ixmp4.conf.settings import Settings

load_dotenv()
# strict typechecking fails due to a bug
# https://docs.pydantic.dev/visual_studio_code/#adding-a-default-with-field
settings = Settings()  # type: ignore



from datetime import datetime
from typing import ClassVar

import pandas as pd

from ixmp4.core.base import BaseFacade, BaseModelFacade
from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import Scenario as ScenarioModel


class Scenario(BaseModelFacade):
    _model: ScenarioModel
    NotFound: ClassVar = ScenarioModel.NotFound
    NotUnique: ClassVar = ScenarioModel.NotUnique

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    @property
    def docs(self):
        try:
            return self.backend.scenarios.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description):
        if description is None:
            self.backend.scenarios.docs.delete(self.id)
        else:
            self.backend.scenarios.docs.set(self.id, description)

    @docs.deleter
    def docs(self):
        try:
            self.backend.scenarios.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Scenario {self.id} name={self.name}>"


class ScenarioRepository(BaseFacade):
    def create(
        self,
        name: str,
    ) -> Scenario:
        model = self.backend.scenarios.create(name)
        return Scenario(_backend=self.backend, _model=model)

    def get(self, name: str) -> Scenario:
        model = self.backend.scenarios.get(name)
        return Scenario(_backend=self.backend, _model=model)

    def list(self, name: str | None = None) -> list[Scenario]:
        scenarios = self.backend.scenarios.list(name=name)
        return [Scenario(_backend=self.backend, _model=s) for s in scenarios]

    def tabulate(self, name: str | None = None) -> pd.DataFrame:
        return self.backend.scenarios.tabulate(name=name)

    def _get_scenario_id(self, scenario: str) -> int | None:
        if scenario is None:
            return None
        elif isinstance(scenario, str):
            obj = self.backend.scenarios.get(scenario)
            return obj.id
        else:
            raise ValueError(f"Invalid reference to scenario: {scenario}")

    def get_docs(self, name: str) -> str | None:
        scenario_id = self._get_scenario_id(name)
        if scenario_id is None:
            return None
        try:
            return self.backend.scenarios.docs.get(dimension_id=scenario_id).description
        except DocsModel.NotFound:
            return None

    def set_docs(self, name: str, description: str | None) -> str | None:
        if description is None:
            self.delete_docs(name=name)
            return None
        scenario_id = self._get_scenario_id(name)
        if scenario_id is None:
            return None
        return self.backend.scenarios.docs.set(
            dimension_id=scenario_id, description=description
        ).description

    def delete_docs(self, name: str) -> None:
        # TODO: this function is failing silently, which we should avoid
        scenario_id = self._get_scenario_id(name)
        if scenario_id is None:
            return None
        try:
            self.backend.scenarios.docs.delete(dimension_id=scenario_id)
            return None
        except DocsModel.NotFound:
            return None



from typing import Any, ClassVar, Generic, Mapping, TypeVar

import pandas as pd
from sqlalchemy.ext.declarative import AbstractConcreteBase
from sqlalchemy.orm.decl_api import declared_attr
from sqlalchemy.orm.exc import NoResultFound

from ixmp4 import db
from ixmp4.data import abstract
from ixmp4.data.db.model import Model
from ixmp4.data.db.run import Run

from ..auth.decorators import guard
from . import base


class TimeSeries(AbstractConcreteBase, base.BaseModel):
    NotFound: ClassVar = abstract.TimeSeries.NotFound
    NotUnique: ClassVar = abstract.TimeSeries.NotUnique
    DeletionPrevented: ClassVar = abstract.TimeSeries.DeletionPrevented

    __abstract__ = True
    parameters: dict = {}

    @declared_attr
    def run__id(cls):
        return db.Column(
            "run__id",
            db.Integer,
            db.ForeignKey("run.id"),
            nullable=False,
            index=True,
        )

    @declared_attr
    def run(cls):
        return db.relationship("Run", backref="time_series", foreign_keys=[cls.run__id])

    @property
    def run_id(self) -> int:
        return self.run__id


ModelType = TypeVar("ModelType", bound=TimeSeries)


class TimeSeriesRepository(
    base.Creator[ModelType],
    base.Retriever[ModelType],
    base.Enumerator[ModelType],
    base.BulkUpserter[ModelType],
    Generic[ModelType],
):
    def join_auth(self, exc: db.sql.Select) -> db.sql.Select:
        if not db.utils.is_joined(exc, Run):
            exc = exc.join(Run, onclause=Run.id == self.model_class.run__id)
        if not db.utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)

        return exc

    def add(self, run_id: int, parameters: Mapping) -> ModelType:
        time_series = self.model_class(run_id=run_id, **parameters)
        self.session.add(time_series)
        return time_series

    @guard("edit")
    def create(self, *args, **kwargs) -> ModelType:
        return super().create(*args, **kwargs)

    @guard("view")
    def get(self, run_id: int, **kwargs: Any) -> ModelType:
        exc = self.select(run={"id": run_id}, **kwargs)

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise self.model_class.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> ModelType:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise self.model_class.NotFound

        return obj

    def select_joined_parameters(self) -> db.sql.Select:
        raise NotImplementedError

    def select(
        self,
        *,
        _exc: db.sql.Select | None = None,
        join_parameters: bool | None = False,
        **kwargs,
    ) -> db.sql.Select:
        if _exc is not None:
            exc = _exc
        elif join_parameters:
            exc = self.select_joined_parameters()
        else:
            exc = db.select(self.model_class)

        return super().select(_exc=exc, **kwargs)

    @guard("view")
    def list(self, *args, **kwargs) -> list[ModelType]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(self, *args, **kwargs) -> pd.DataFrame:
        return super().tabulate(*args, **kwargs)

    @guard("edit")
    def bulk_upsert(self, df: pd.DataFrame) -> None:
        return super().bulk_upsert(df)



from fastapi import APIRouter, Body, Depends, Query
from pydantic import StrictInt, StrictStr

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.optimization.indexset.filter import OptimizationIndexSetFilter

from .. import deps
from ..base import BaseModel, EnumerationOutput, Pagination
from ..decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/indexsets",
    tags=["optimization", "indexsets"],
)


class IndexSetInput(BaseModel):
    run_id: int
    name: str


class ElementsInput(BaseModel):
    elements: StrictInt | list[StrictInt | StrictStr] | StrictStr


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.IndexSet])
def query(
    filter: OptimizationIndexSetFilter = Body(OptimizationIndexSetFilter()),
    table: bool = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.optimization.indexsets.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.optimization.indexsets.count(_filter=filter),
        pagination=pagination,
    )


@autodoc
@router.post("/", response_model=api.IndexSet)
def create(
    indexset: IndexSetInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.indexsets.create(**indexset.model_dump())


@autodoc
@router.patch("/{indexset_id}/")
def add_elements(
    indexset_id: int,
    elements: ElementsInput,
    backend: Backend = Depends(deps.get_backend),
):
    backend.optimization.indexsets.add_elements(
        indexset_id=indexset_id, **elements.model_dump()
    )
